/*---------------------------------------------------------------------------

  FILENAME:
        qms-analog.h

  PURPOSE:
        Provide a qmail-scanner log analysis tool.

  REVISION HISTORY:
        Date            Engineer        Revision        Remarks
        05/29/04        M.S. Teel       0               Original

  NOTES:
        File records passed in as stdin and results written to stdout to 
        allow piping.
        
        This requires qmail-scanner-1.22, clamAV, Spamassassin, and the
        included patch qmail-scanner-1.22-st-qms-YYYYMMDD.patch applied to 
        a clean .../qmail-scanner-1.22 distribution and successfully
        configured.
        
        This patch adds support for a new (legible) event log file
        /var/spool/qmailscan/qms-events.log. This file is required as 
        input to this utility.

  LICENSE:
        Copyright (c) 2004, Mark S. Teel (mark@teel.ws)
  
        This source code is released for free distribution under the terms 
        of the GNU General Public License.
  
----------------------------------------------------------------------------*/
#ifndef INC_qmsanalogh
#define INC_qmsanalogh

/*  ... System include files
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

/*  ... Library include files
*/

/*  ... Local include files
*/
#include "session.h"


/*  ... this one is dependent upon the size of the array below
*/
#define TYPE_MAX                    39


#define MAX_TOKENS          6

enum _sort_keys
{
    KEY_ALPHA       = 0,
    KEY_MSGBW,
    KEY_VIRUS,
    KEY_SAAVG,
    KEY_SADET,
    KEY_MAX_KEY
};




typedef enum _display_type
{
    DISPLAY_NEVER       = 0,
    DISPLAY_ALWAYS      = 1,
    DISPLAY_NON_ZERO    = 2
} DISPLAY_TYPE;


typedef struct _stats_type_tag
{
    char        *matchStr;
    int         average;
    int         display;
    char        *description;
} STAT_DATA;


struct _qms_log_data
{
    double          stats[TYPE_MAX];
    unsigned long   dpoints[TYPE_MAX];
};


/*  ... define the range that defines a virus detection
    ... WARNING: all virus types must be contiguous in the list!
*/
#define STAT_VIRUS_FIRST        10
#define STAT_VIRUS_LAST         34


#ifdef QMS_INCLUDE_TYPE_ARRAY
STAT_DATA qmsStats[TYPE_MAX] =
{
    {"CONNECT-SMTP",0,DISPLAY_ALWAYS,                "External SMTP sessions:                  "},
    {"CONNECT-PIPE",0,DISPLAY_ALWAYS,                "Internal SMTP sessions:                  "},
    {"SMTP-DROP",0,DISPLAY_ALWAYS,                   "SMTP msgs dropped:                       "},
    {"ERROR",0,DISPLAY_ALWAYS,                       "Scan Errors:                             "},
    {"TYPE:PLAIN",0,DISPLAY_ALWAYS,                  "Non-MIME or text/plain (msgs):           "},
    {"TYPE:MIXED",0,DISPLAY_ALWAYS,                  "MIME (msgs):                             "},
    {"SCANTIME",1,DISPLAY_ALWAYS,                    "Average scan time/msg (secs):            "},
    {"HEADER",0,DISPLAY_NEVER,                       "NA"},
    {"ENV-HEADER",0,DISPLAY_NEVER,                   "NA"},
    {"SPAM-RESULT",0,DISPLAY_NEVER,                  "NA"},
    {"QMSWC:BAD_HDR_CHARS",0,DISPLAY_ALWAYS,         "QMS WC bad char (msgs):                  "},
    {"QMSWC:BAD_HDR_BREAKAGE",0,DISPLAY_ALWAYS,      "QMS WC header breakage (msgs):           "},
    {"QMSWC:BAD_HDR_MIME",0,DISPLAY_ALWAYS,          "QMS WC bad MIME hdr (msgs):              "},
    {"QMSWC:BAD_MIME_CONTENT",0,DISPLAY_ALWAYS,      "QMS WC bad MIME content (msgs):          "},
    {"QMSWC:BAD_MIME_FILENAME",0,DISPLAY_ALWAYS,     "QMS WC bad MIME filename (msgs):         "},
    {"QMSWC:BAD_MIME_BOUNDARY",0,DISPLAY_ALWAYS,     "QMS WC bad MIME boundary (msgs):         "},
    {"QMSWC:BAD_MIME_ASSOCIATION",0,DISPLAY_ALWAYS,  "QMS WC bad MIME assoc (msgs):            "},
    {"QMSWC:BAD_MIME_WINBLOWS",0,DISPLAY_ALWAYS,     "QMS WC bad MIME windows file (msgs):     "},
    {"QMSWC:BAD_MIME_ZIP",0,DISPLAY_ALWAYS,          "QMS WC bad zip file (msgs):              "},
    {"PERLSCAN:BAD_MIME_HEADER",0,DISPLAY_ALWAYS,    "PERLSCAN bad MIME hdr (msgs):            "},
    {"PERLSCAN:BAD_HDR_DB",0,DISPLAY_ALWAYS,         "PERLSCAN bad hdr found in db (msgs):     "},
    {"PERLSCAN:BAD_ATTACH_LENGTH",0,DISPLAY_ALWAYS,  "PERLSCAN bad attach length (msgs):       "},
    {"PERLSCAN:BAD_ATTACHMENT_TYPE",0,DISPLAY_ALWAYS,"PERLSCAN bad attachment type (msgs):     "},
    {"UNZIP:PASSWORD_PROTECTED",0,DISPLAY_ALWAYS,    "UNZIP password protected (msgs):         "},
    {"CLAMAV",0,DISPLAY_NON_ZERO,                    "CLAMAV virus found (msgs):               "},
    {"AVPAV",0,DISPLAY_NON_ZERO,                     "AVPLinux virus found (msgs):             "},
    {"CSAV",0,DISPLAY_NON_ZERO,                      "Command scanner virus found (msgs):      "},
    {"FPROTAV",0,DISPLAY_NON_ZERO,                   "F-prot scanner virus found (msgs):       "},
    {"FSECUREAV",0,DISPLAY_NON_ZERO,                 "Fsecure virus found (msgs):              "},
    {"HBEDVAV",0,DISPLAY_NON_ZERO,                   "H+BEDV virus found (msgs):               "},
    {"INNOCAV",0,DISPLAY_NON_ZERO,                   "Innoculan virus found (msgs):            "},
    {"ISCANAV",0,DISPLAY_NON_ZERO,                   "Iscan virus found (msgs):                "},
    {"RAVLINAV",0,DISPLAY_NON_ZERO,                  "Ravlin virus found (msgs):               "},
    {"UVSCANAV",0,DISPLAY_NON_ZERO,                  "McAfee virus found (msgs):               "},
    {"VEXIRAV",0,DISPLAY_NON_ZERO,                   "Vexira virus found (msgs):               "},
    {"SPAM-DETECT:DELETE",0,DISPLAY_ALWAYS,          "SPAM deleted (msgs):                     "},
    {"SPAM-DETECT:MARK",0,DISPLAY_ALWAYS,            "SPAM detected (msgs):                    "},
    {"SPAM-DETECT:QUARANTINE",0,DISPLAY_ALWAYS,      "SPAM quarantined (msgs):                 "},
    {"SPAM-DETECT:REJECT",0,DISPLAY_ALWAYS,          "SPAM rejected (msgs):                    "}
};
#else
extern STAT_DATA qmsStats[];
#endif

#endif

