/*---------------------------------------------------------------------------

  FILENAME:
        session.c

  PURPOSE:
        Provide session processing for the qmail-scanner log analysis tool.

  REVISION HISTORY:
        Date            Engineer        Revision        Remarks
        05/29/04        M.S. Teel       0               Original

  NOTES:
        
  LICENSE:
        Copyright (c) 2004, Mark S. Teel (mark@teel.ws)
  
        This source code is released for free distribution under the terms 
        of the GNU General Public License.
  
----------------------------------------------------------------------------*/

/*  ... System include files
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <ctype.h>

/*  ... Library include files
*/

/*  ... Local include files
*/
#include "qms-analog.h"

/*  ... global memory declarations
*/

/*  ... global memory referenced
*/

/*  ... static (local) memory declarations
*/
static QMS_SESSION_MGR  sessionData;


/*  ... static functions
*/
static void insertAlpha (QMS_ACCOUNT *account)
{
    QMS_ACCOUNT     *tempacct;

    for (tempacct = (QMS_ACCOUNT *)radListGetFirst (&sessionData.accountList);
         tempacct != NULL;
         tempacct = (QMS_ACCOUNT *)radListGetNext (&sessionData.accountList,
                                                   (NODE_PTR)tempacct))
    {
        if (strcmp (account->accountName, tempacct->accountName) < 0)
        {
            radListInsertBefore (&sessionData.accountList,
                                 (NODE_PTR)tempacct,
                                 (NODE_PTR)account);
            return;
        }
    }

    /* if here, we have to add it to the end */
    radListAddToEnd (&sessionData.accountList, (NODE_PTR)account);
    return;
}

static void insertMsgBW (QMS_ACCOUNT *account)
{
    QMS_ACCOUNT     *tempacct;
    double          acctBW, tempBW;

    acctBW = account->stats[ACCT_DELIVER] + account->stats[ACCT_RELAY];

    for (tempacct = (QMS_ACCOUNT *)radListGetFirst (&sessionData.accountList);
         tempacct != NULL;
         tempacct = (QMS_ACCOUNT *)radListGetNext (&sessionData.accountList,
                                                   (NODE_PTR)tempacct))
    {
        tempBW = tempacct->stats[ACCT_DELIVER] + tempacct->stats[ACCT_RELAY];
        if (acctBW > tempBW)
        {
            radListInsertBefore (&sessionData.accountList,
                                 (NODE_PTR)tempacct,
                                 (NODE_PTR)account);
            return;
        }
    }

    /* if here, we have to add it to the end */
    radListAddToEnd (&sessionData.accountList, (NODE_PTR)account);
    return;
}

static void insertVirus (QMS_ACCOUNT *account)
{
    QMS_ACCOUNT     *tempacct;

    for (tempacct = (QMS_ACCOUNT *)radListGetFirst (&sessionData.accountList);
         tempacct != NULL;
         tempacct = (QMS_ACCOUNT *)radListGetNext (&sessionData.accountList,
                                                   (NODE_PTR)tempacct))
    {
        if (account->stats[ACCT_VIRUS_IN] > tempacct->stats[ACCT_VIRUS_IN])
        {
            radListInsertBefore (&sessionData.accountList,
                                 (NODE_PTR)tempacct,
                                 (NODE_PTR)account);
            return;
        }
    }

    /* if here, we have to add it to the end */
    radListAddToEnd (&sessionData.accountList, (NODE_PTR)account);
    return;
}

static void insertSaAvg (QMS_ACCOUNT *account)
{
    QMS_ACCOUNT     *tempacct;
    double          acctAvg, tempAvg;

    if (account->stats[ACCT_SP_COUNT] > 0)
        acctAvg = account->stats[ACCT_SP_SCORE]/account->stats[ACCT_SP_COUNT];
    else
        acctAvg = 0;

    for (tempacct = (QMS_ACCOUNT *)radListGetFirst (&sessionData.accountList);
         tempacct != NULL;
         tempacct = (QMS_ACCOUNT *)radListGetNext (&sessionData.accountList,
                                                   (NODE_PTR)tempacct))
    {
        if (tempacct->stats[ACCT_SP_COUNT] > 0)
            tempAvg = tempacct->stats[ACCT_SP_SCORE]/tempacct->stats[ACCT_SP_COUNT];
        else
            tempAvg = 0;
        if (acctAvg > tempAvg)
        {
            radListInsertBefore (&sessionData.accountList,
                                 (NODE_PTR)tempacct,
                                 (NODE_PTR)account);
            return;
        }
    }

    /* if here, we have to add it to the end */
    radListAddToEnd (&sessionData.accountList, (NODE_PTR)account);
    return;
}

static void insertSaDet (QMS_ACCOUNT *account)
{
    QMS_ACCOUNT     *tempacct;
    double          acctNum, tempNum;

    acctNum = account->stats[ACCT_SP_MARK] +
              account->stats[ACCT_SP_DELETE] +
              account->stats[ACCT_SP_REJECT] +
              account->stats[ACCT_SP_QUARANTINE];

    for (tempacct = (QMS_ACCOUNT *)radListGetFirst (&sessionData.accountList);
         tempacct != NULL;
         tempacct = (QMS_ACCOUNT *)radListGetNext (&sessionData.accountList,
                                                   (NODE_PTR)tempacct))
    {
        tempNum = tempacct->stats[ACCT_SP_MARK] +
                  tempacct->stats[ACCT_SP_DELETE] +
                  tempacct->stats[ACCT_SP_REJECT] +
                  tempacct->stats[ACCT_SP_QUARANTINE];
        if (acctNum > tempNum)
        {
            radListInsertBefore (&sessionData.accountList,
                                 (NODE_PTR)tempacct,
                                 (NODE_PTR)account);
            return;
        }
    }

    /* if here, we have to add it to the end */
    radListAddToEnd (&sessionData.accountList, (NODE_PTR)account);
    return;
}
/*  ... jump table for sort functions
*/
static void (*sortFuncs[KEY_MAX_KEY]) (QMS_ACCOUNT *account) =
{
    insertAlpha,
    insertMsgBW,
    insertVirus,
    insertSaAvg,
    insertSaDet
};


static QMS_SESSION *getSession (int pid)
{
    QMS_SESSION     *session;
    
    for (session = (QMS_SESSION *)radListGetFirst (&sessionData.sessionList);
         session != NULL;
         session = (QMS_SESSION *)radListGetNext (&sessionData.sessionList,
                                                  (NODE_PTR)session))
    {
        if (session->procId == pid)
            return session;
    }
    
    return NULL;
}

static QMS_ACCOUNT *getAccount (QMS_SESSION *session)
{
    QMS_ACCOUNT     *account, *tempacct;
    
    for (account = (QMS_ACCOUNT *)radListGetFirst (&sessionData.accountList);
         account != NULL;
         account = (QMS_ACCOUNT *)radListGetNext (&sessionData.accountList,
                                                  (NODE_PTR)account))
    {
        if (!strcmp (account->accountName, session->accountName))
            return account;
    }

    /* if we are here, we need to create a new one */
    account = (QMS_ACCOUNT *)malloc (sizeof (QMS_ACCOUNT));
    if (account == NULL)
        return NULL;
    
    memset (account, 0, sizeof (*account));
    
    strcpy (account->accountName, session->accountName);
    
    /* insert in list */
    radListAddToEnd (&sessionData.accountList, (NODE_PTR)account);

    return account;
}

static void sortAccounts (void)
{
    LIST            tempList;
    QMS_ACCOUNT     *tempAcct;

    radListReset (&tempList);

    for (tempAcct = (QMS_ACCOUNT *)radListRemoveFirst (&sessionData.accountList);
         tempAcct != NULL;
         tempAcct = (QMS_ACCOUNT *)radListRemoveFirst (&sessionData.accountList))
    {
        radListAddToEnd (&tempList, (NODE_PTR)tempAcct);
    }

    for (tempAcct = (QMS_ACCOUNT *)radListRemoveFirst (&tempList);
         tempAcct != NULL;
         tempAcct = (QMS_ACCOUNT *)radListRemoveFirst (&tempList))
    {
        (*sortFuncs[sessionData.sortKey]) (tempAcct);
    }
}


static QMS_VIRUS_DATA *getVirusData (char *descr)
{
    QMS_VIRUS_DATA  *virusData;
    
    for (virusData = (QMS_VIRUS_DATA *)radListGetFirst (&sessionData.virusList);
         virusData != NULL;
         virusData = (QMS_VIRUS_DATA *)radListGetNext (&sessionData.virusList,
                                                  (NODE_PTR)virusData))
    {
        if (!strcmp (virusData->description, descr))
            return virusData;
    }
    
    /* if we are here, we need to create a new one */
    virusData = (QMS_VIRUS_DATA *)malloc (sizeof (QMS_VIRUS_DATA));
    if (virusData == NULL)
        return NULL;
    
    memset (virusData, 0, sizeof (*virusData));
    
    strcpy (virusData->description, descr);
    
    radListAddToEnd (&sessionData.virusList, (NODE_PTR)virusData);
    
    return virusData;
}

static void sortViruses (void)
{
    LIST            tempList;
    QMS_VIRUS_DATA  *tempVirus, *tempVirus1;

    radListReset (&tempList);

    for (tempVirus = (QMS_VIRUS_DATA *)radListRemoveFirst (&sessionData.virusList);
         tempVirus != NULL;
         tempVirus = (QMS_VIRUS_DATA *)radListRemoveFirst (&sessionData.virusList))
    {
        for (tempVirus1 = (QMS_VIRUS_DATA *)radListGetFirst (&tempList);
             tempVirus1 != NULL;
             tempVirus1 = (QMS_VIRUS_DATA *)radListGetNext (&tempList, (NODE_PTR)tempVirus1))
        {
            if (tempVirus->count > tempVirus1->count)
            {
                radListInsertBefore (&tempList, (NODE_PTR)tempVirus1, (NODE_PTR)tempVirus);
                break;
            }
        }
        if (tempVirus1 == NULL)
            radListAddToEnd (&tempList, (NODE_PTR)tempVirus);
    }

    for (tempVirus = (QMS_VIRUS_DATA *)radListRemoveFirst (&tempList);
         tempVirus != NULL;
         tempVirus = (QMS_VIRUS_DATA *)radListRemoveFirst (&tempList))
    {
        radListAddToEnd (&sessionData.virusList, (NODE_PTR)tempVirus);
    }
}




int qmsSessionInit (int sortKey)
{
    radListReset (&sessionData.sessionList);
    radListReset (&sessionData.accountList);
    radListReset (&sessionData.virusList);
    sessionData.sortKey = sortKey;
    
    return 0;
}

void qmsSessionExit (void)
{
    NODE_PTR    nptr;
    
    for (nptr = radListGetFirst (&sessionData.sessionList);
         nptr != NULL;
         nptr = radListGetFirst (&sessionData.sessionList))
    {
        radListRemove (&sessionData.sessionList, nptr);
        free (nptr);
    }
    
    for (nptr = radListGetFirst (&sessionData.accountList);
         nptr != NULL;
         nptr = radListGetFirst (&sessionData.accountList))
    {
        radListRemove (&sessionData.accountList, nptr);
        free (nptr);
    }
    
    for (nptr = radListGetFirst (&sessionData.virusList);
         nptr != NULL;
         nptr = radListGetFirst (&sessionData.virusList))
    {
        radListRemove (&sessionData.virusList, nptr);
        free (nptr);
    }
    
    return;
}

void qmsSessionProcess (char *tokens[])
{
    int             i, j, pid = atoi (tokens[1]);
    QMS_SESSION     *session;
    double          scantime;
    QMS_ACCOUNT     *account;
    char            localDomains[2048];
    char            strtokLine[2048];
    char            *strtokPtr = strtokLine;
    char            *domain;
    char            srcName[128];
    char            srcDomain[128];
    QMS_DOMAIN      destDomains[MAX_DEST_DOMAINS];
    int             lastDestDomainIndex;
    QMS_VIRUS_DATA  *virusData;
    
    if (!strncmp (tokens[2], "CONNECT", 7))
    {
        /* new session - create an instance */
        session = (QMS_SESSION *) malloc (sizeof (QMS_SESSION));
        if (session == NULL)
        {
            printf ("session: ERROR: allocation error for pid %d\n", pid);
            return;
        }
        memset (session, 0, sizeof (*session));
        
        session->procId = pid;
        
        /* add to our session list */
        radListAddToEnd (&sessionData.sessionList, (NODE_PTR)session);
        
        return;
    }
    else if (!strcmp (tokens[2], "ENV-HEADER"))
    {
        /* let's get the account name */
        session = getSession (pid);
        if (session == NULL)
            return;
        
        if (tokens[3] == NULL || tokens[4] == NULL || tokens[5] == NULL)
        {
            /* it's missing some data - delete the session */
            radListRemove (&sessionData.sessionList, (NODE_PTR)session);
            free (session);
            return;
        }
        
        /* strip any single quotes */
        for (i = 0, j = 0; i < strlen (tokens[3]); i ++)
        {
            if (tokens[3][i] == 0x27)
                continue;
            
            localDomains[j] = tolower (tokens[3][i]);   // not case sensitive
            j ++;
        }
        localDomains[j] = 0;
        
        /* get our src and dest address domains */
        for (i = 0; tokens[4][i] != 0; i ++)
        {
            tokens[4][i] = tolower (tokens[4][i]);
        }
        for (i = 0; tokens[4][i] != 0; i ++)
        {
            if (tokens[4][i] == '@')
            {
                i ++;
                break;
            }
        }
        strcpy (srcDomain, &tokens[4][i]);
        strcpy (srcName, tokens[4]);
        
        /* now lets look at each domain in the dest list */
        lastDestDomainIndex = 0;
        
        for (domain = strtok_r (tokens[5], ",\n", &strtokPtr);
             domain != NULL;
             domain = strtok_r (NULL, ",\n", &strtokPtr))
        {
            for (i = 0; domain[i] != 0; i ++)
            {
                domain[i] = tolower (domain[i]);
            }
            for (i = 0; domain[i] != 0; i ++)
            {
                if (domain[i] == '@')
                {
                    i ++;
                    break;
                }
            }
            if (domain[i] == 0)
                continue;
            
            if (lastDestDomainIndex >= MAX_DEST_DOMAINS)
                break;
            
            strcpy (destDomains[lastDestDomainIndex].domain, &domain[i]);
            strcpy (destDomains[lastDestDomainIndex].name, domain);
            lastDestDomainIndex ++;
        }
        
        /* now lets look at each domain in the domain list */
        for (domain = strtok_r (localDomains, ",", &strtokPtr);
             domain != NULL;
             domain = strtok_r (NULL, ",", &strtokPtr))
        {
            /* compare the source domain first */
            if (!strcmp (domain, srcDomain))
            {
                /* we have a match - sender - save the account name */
                session->incoming = 0;
                if (session->stats[ACCT_VIRUS_OUT] || session->stats[ACCT_VIRUS_IN])
                {
                    session->stats[ACCT_VIRUS_OUT] = 1;
                    session->stats[ACCT_VIRUS_IN] = 0;
                }
                else
                    session->stats[ACCT_RELAY] = 1;

                strcpy (session->accountName, srcName);
                return;
            }

            /* now look at the dest domains */
            for (i = 0; i < lastDestDomainIndex; i ++)
            {
                if (!strcmp (domain, destDomains[i].domain))
                {
                    /* we have a match - receiver - save the account name */
                    session->incoming = 1;
                    if (session->stats[ACCT_VIRUS_OUT] || session->stats[ACCT_VIRUS_IN])
                    {
                        session->stats[ACCT_VIRUS_OUT] = 0;
                        session->stats[ACCT_VIRUS_IN] = 1;
                    }
                    else
                        session->stats[ACCT_DELIVER] = 1;

                    strcpy (session->accountName, destDomains[i].name);
                    return;
                }
            }
        }
            
        /* if we are here, we found no local domain matches - punt */
        /* delete the session */
        radListRemove (&sessionData.sessionList, (NODE_PTR)session);
        free (session);
        return;
    }
    else if (!strcmp (tokens[2], "SMTP-DROP"))
    {
        /* remove the session, SMTP is giving up */
        session = getSession (pid);
        if (session == NULL)
            return;
        
        radListRemove (&sessionData.sessionList, (NODE_PTR)session);
        free (session);
        
        return;
    }
    else if (!strcmp (tokens[2], "SCANTIME"))
    {
        /* get the scan time and close out the session */
        scantime = strtod (tokens[3], NULL);
        
        session = getSession (pid);
        if (session == NULL)
            return;
        
        /* ... now save to the account ... */
        if (strlen (session->accountName) > 0)
        {
            session->stats[ACCT_TIME_TOTAL] = scantime;
        
            account = getAccount (session);
            if (account == NULL)
            {
                printf ("session: ERROR: account error for pid %d\n", pid);
                return;
            }
                
            for (i = 0; i < ACCT_MAX; i ++)
            {
                account->stats[i] += session->stats[i];
            }
        }
        
        /* delete the session */
        radListRemove (&sessionData.sessionList, (NODE_PTR)session);
        free (session);
        
        return;
    }
    else if (!strcmp (tokens[2], "SPAM-RESULT"))
    {
        /* spam result record - get the score */
        session = getSession (pid);
        if (session == NULL)
            return;
        
        session->stats[ACCT_SP_SCORE] += strtod (tokens[3], NULL);
        session->stats[ACCT_SP_COUNT] ++;
        
        return;
    }
    else if (!strcmp (tokens[2], "SPAM-DETECT"))
    {
        /* spam detect record - bump the right stat */
        session = getSession (pid);
        if (session == NULL)
        {
            return;
        }
        if (!strcmp (tokens[3], "MARK"))
        {
            session->stats[ACCT_SP_MARK] = 1; 
        }
        else if (!strcmp (tokens[3], "DELETE"))
        {
            session->stats[ACCT_DELIVER] = 0; 
            session->stats[ACCT_SP_DELETE] = 1; 
        }
        else if (!strcmp (tokens[3], "REJECT"))
        {
            session->stats[ACCT_DELIVER] = 0; 
            session->stats[ACCT_SP_REJECT] = 1; 
        }
        else if (!strcmp (tokens[3], "QUARANTINE"))
        {
            session->stats[ACCT_DELIVER] = 0; 
            session->stats[ACCT_SP_QUARANTINE] = 1; 
        }
        
        return;
    }
    else
    {
        /* look for a matching virus record */
        
        /* is this a virus record? */
        for (i = STAT_VIRUS_FIRST; i <= STAT_VIRUS_LAST; i ++)
        {
            if (!strncmp (tokens[2], qmsStats[i].matchStr, strlen (tokens[2])))
            {
                session = getSession (pid);
                if (session == NULL)
                {
                    return;
                }
                /* match! */
                if (session->incoming)
                {
                    session->stats[ACCT_VIRUS_IN] = 1;
                    session->stats[ACCT_DELIVER] = 0;
                }
                else
                {
                    session->stats[ACCT_VIRUS_OUT] = 1;
                    session->stats[ACCT_RELAY] = 0;
                }
                
                if (tokens[3] == NULL)
                    return;
                
                /* also store it in the virus list */
                virusData = getVirusData (tokens[3]);
                if (virusData == NULL)
                {
                    printf ("session: ERROR: virus data alloc error - pid %d",
                            pid);
                    return;
                }
    
                virusData->count ++;
                return;
            }
        }
    }
    
    return;
}

void qmsSessionPrintResults (void)
{
    QMS_VIRUS_DATA      *virusData;
    QMS_ACCOUNT         *account;
    double              scoreAvg, msgPercent, totalSuccesses = 0;
    
    
    /* sort the viruses first */
    sortViruses ();
    
    printf ("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    printf ("==> Detected Viruses Report\n\n");
    printf ("    Count  Description\n");
    printf ("------------------------------------------------------------------------------\n");
    
    for (virusData = (QMS_VIRUS_DATA *)radListGetFirst (&sessionData.virusList);
         virusData != NULL;
         virusData = (QMS_VIRUS_DATA *)radListGetNext (&sessionData.virusList,
                                                       (NODE_PTR)virusData))
    {
        printf ("  %7d  %s\n", virusData->count, virusData->description);
    }
    printf ("\n");
    

    /* sum up a few things */
    for (account = (QMS_ACCOUNT *)radListGetFirst (&sessionData.accountList);
         account != NULL;
         account = (QMS_ACCOUNT *)radListGetNext (&sessionData.accountList,
                                                       (NODE_PTR)account))
    {
        totalSuccesses += account->stats[ACCT_DELIVER];
        totalSuccesses += account->stats[ACCT_RELAY];
    }

    /* sort the accounts based on user input */
    sortAccounts ();

    printf ("\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    printf ("==> Local Accounts Report\n\n");
    
    printf ("MsgRx MsgTx %Total ScanTime VirusRx VirusTx SA-AVG SA-MRK SA-DEL SA-REJ SA-QUA\n");
    printf ("------------------------------------------------------------------------------\n");
    for (account = (QMS_ACCOUNT *)radListGetFirst (&sessionData.accountList);
         account != NULL;
         account = (QMS_ACCOUNT *)radListGetNext (&sessionData.accountList,
                                                       (NODE_PTR)account))
    {
        printf ("%s:\n", account->accountName);
        
        if (totalSuccesses > 0)
            msgPercent = 100 * 
                         ((account->stats[ACCT_DELIVER]+account->stats[ACCT_RELAY])/totalSuccesses);
        else
            msgPercent = 0;
        
        if (account->stats[ACCT_SP_COUNT])
            scoreAvg = account->stats[ACCT_SP_SCORE]/account->stats[ACCT_SP_COUNT];
        else
            scoreAvg = 0.0;
        
        printf ("%5d %5d  %5.1f %8.1f %7d %7d   %4.1f %6d %6d %6d %6d\n",
                (int)account->stats[ACCT_DELIVER],
                (int)account->stats[ACCT_RELAY],
                msgPercent,
                account->stats[ACCT_TIME_TOTAL],
                (int)account->stats[ACCT_VIRUS_IN],
                (int)account->stats[ACCT_VIRUS_OUT],
                scoreAvg,
                (int)account->stats[ACCT_SP_MARK],
                (int)account->stats[ACCT_SP_DELETE],
                (int)account->stats[ACCT_SP_REJECT],
                (int)account->stats[ACCT_SP_QUARANTINE]);
                
        printf ("\n");
    }
    
    return;
}

