/*---------------------------------------------------------------------------

  FILENAME:
        qms-analog.c

  PURPOSE:
        Provide a qmail-scanner log analysis tool.

  REVISION HISTORY:
        Date            Engineer        Revision        Remarks
        05/21/04        M.S. Teel       0               Original

  NOTES:
        File records passed in as stdin and results written to stdout to 
        allow piping.
        
        This requires qmail-scanner-1.22, clamAV, Spamassassin, and the
        included patch qmail-scanner-1.22-st-qms-YYYYMMDD.patch applied to 
        a clean .../qmail-scanner-1.22 distribution and successfully
        configured.
        
        This patch adds support for a new (legible) event log file
        /var/spool/qmailscan/qms-events.log. This file is required as 
        input to this utility.

  LICENSE:
        Copyright (c) 2004, Mark S. Teel (mark@teel.ws)
  
        This source code is released for free distribution under the terms 
        of the GNU General Public License.
  
----------------------------------------------------------------------------*/

/*  ... System include files
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

/*  ... Library include files
*/

/*  ... Local include files
*/
#define QMS_INCLUDE_TYPE_ARRAY 1
#include "qms-analog.h"

/*  ... global memory declarations
*/

/*  ... global memory referenced
*/

/*  ... static (local) memory declarations
*/

    
static struct _qms_log_data QMS_DATA;


    
/*	... methods
*/

static void usage (void)
{
    printf ("qms-analog - Qmail-Scanner Log Analyzer\n\n");
    printf ("usage: qms-analog hours-of-history <sort-key>\n\n");
    printf ("       hours-of-history        (0 - n) hours of history to collect\n");
    printf ("                               0 => all records\n");
    printf ("       sort-key (optional)     sort key for account statistics\n");
    printf ("           msgbw               (default) msg bandwidth - successful msgs\n");
    printf ("           alpha               alphanumeric by account name\n");
    printf ("           virus               number of viruses received\n");
    printf ("           saavg               Spamassassin avg score\n");
    printf ("           sadet               Spamassassin msgs detected\n\n");

    return;
}

#define QMS_INLINE_MAX		0x8000

/*  ... the entry point
*/
int main (int argc, char *argv[])
{
    char        inLine[QMS_INLINE_MAX], strtokLine[QMS_INLINE_MAX];
    char        *strtokPtr = strtokLine;
    int         i, j, length, found, numTokens = 0;
    int         qmsStrLens[TYPE_MAX];
    time_t      firstDate, lastDate, startTime, currTime;
    char        *tokens[MAX_TOKENS];
    char        tempToken[64];
    int         hours, sortKey = KEY_MSGBW;
    
    if (argc < 2)
    {
        usage ();
        return 1;
    }

    /* figure out the start time for records */
    hours = atoi (argv[1]);
    if (hours <= 0)
    {
        startTime = 0;
    }
    else
    {
        startTime = time (NULL);
        startTime -= hours * 3600;
    }

    /* initialize our sort key */
    if (argc > 2)
    {
        if (!strcmp (argv[2], "alpha"))
            sortKey = KEY_ALPHA;
        else if (!strcmp (argv[2], "msgbw"))
            sortKey = KEY_MSGBW;
        else if (!strcmp (argv[2], "virus"))
            sortKey = KEY_VIRUS;
        else if (!strcmp (argv[2], "saavg"))
            sortKey = KEY_SAAVG;
        else if (!strcmp (argv[2], "sadet"))
            sortKey = KEY_SADET;
    }

    /* init the session mgr */
    qmsSessionInit (sortKey);
    
    memset (&QMS_DATA.stats, 0, TYPE_MAX * sizeof (QMS_DATA.stats[0]));    
    memset (&QMS_DATA.dpoints, 0, TYPE_MAX * sizeof (QMS_DATA.dpoints[0]));
    for (i = 0; i < MAX_TOKENS; i ++)
        tokens[i] = 0;

    for (i = 0; i < TYPE_MAX; i ++)
    {
        qmsStrLens[i] = strlen (qmsStats[i].matchStr);
    }
    
    firstDate = lastDate = 0;
    
    
    /* read input from stdin */
    while (fgets (inLine, QMS_INLINE_MAX, stdin) != NULL)
    {
/*      ... process the record
*/
        if ((length = strlen (inLine)) == 0)
            continue;
        
        /* weed out some silliness */
        for (i = 0, j = 0; i < length; i ++)
        {
            if (inLine[i] == ':')
                j ++;
            if (j == 2)
            {
                i ++;
                break;
            }
        }

        if (i == length ||
            inLine[i] == ' ' ||
            inLine[i] == '-' ||
            !strncmp (&inLine[i], "HEADER", 6))
        {
            continue;
        }
        
    
//printf ("DUMPLN: %s\n", inLine);

        /* parse out the tokens */
        if ((tokens[0] = strtok_r (inLine, ":\n", &strtokPtr)) == NULL)
            continue;
        for (i = 1; i < MAX_TOKENS; i ++)
        {
            if ((tokens[i] = strtok_r (NULL, ":\n", &strtokPtr)) == NULL)
            {
                numTokens = i;
                break;
            }
        }
        
        /* is this record to be included? */
        currTime = (time_t)atol (tokens[0]);
        if (currTime < startTime)
            continue;
        
        /* do session processing here ... */
        qmsSessionProcess (tokens);
        
        
        /* deal with special cases */
        if (!strcmp (tokens[2], "TYPE"))
        {
            sprintf (tempToken, "%s:%s", tokens[2], tokens[3]);
            tokens[2] = tempToken;
            tokens[3] = NULL;
        }
        else if (!strcmp (tokens[2], "QMSWC"))
        {
            sprintf (tempToken, "%s:%s", tokens[2], tokens[3]);
            tokens[2] = tempToken;
            tokens[3] = NULL;
        }
        else if (!strcmp (tokens[2], "PERLSCAN"))
        {
            sprintf (tempToken, "%s:%s", tokens[2], tokens[3]);
            tokens[2] = tempToken;
            tokens[3] = NULL;
        }
        else if (!strcmp (tokens[2], "UNZIP"))
        {
            sprintf (tempToken, "%s:%s", tokens[2], tokens[3]);
            tokens[2] = tempToken;
            tokens[3] = NULL;
        }
        else if (!strcmp (tokens[2], "SPAM-DETECT"))
        {
            sprintf (tempToken, "%s:%s", tokens[2], tokens[3]);
            tokens[2] = tempToken;
            tokens[3] = NULL;
        }
        
        
        if (firstDate == 0)
            firstDate = currTime;
        lastDate = currTime;
            
        found = 0;
        
        /* determine what log type it is */
        for (i = 0; !found && i < TYPE_MAX; i ++)
        {
            if (strncmp (tokens[2], qmsStats[i].matchStr, qmsStrLens[i]) == 0)
            {
                /* we have a match! */
                if (qmsStats[i].average)
                {
                    QMS_DATA.stats[i] += strtod (tokens[3], (char **)NULL);
                    QMS_DATA.dpoints[i] ++;
                }
                else
                {
                    QMS_DATA.stats[i] ++;
                }
                    
                found = 1;
            }
        }
    } 

    /* write output to stdout */
    printf ("qms-analog: Qmail-Scanner Log File Analysis\n\n");
    ctime_r (&firstDate, inLine);
    printf ("For records beginning: %s", inLine);
    ctime_r (&lastDate, inLine);
    printf ("               ending: %s\n\n", inLine);
    
    printf ("Note: QMS WC stats are qmail-scanner working\n");
    printf ("      copy generated stats.\n\n\n");
        
    printf ("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    printf ("==> Server Report\n\n");
    for (i = 0; i < TYPE_MAX; i ++)
    {
        if (qmsStats[i].display == DISPLAY_NEVER)
            continue;
        
        if ((qmsStats[i].display == DISPLAY_NON_ZERO) && (QMS_DATA.stats[i] == 0.0))
            continue;
            
            
        if (qmsStats[i].average)
        {
            if (QMS_DATA.dpoints[i] > 0)
                printf ("    %s%g\n",  
                        qmsStats[i].description, 
                        QMS_DATA.stats[i]/(double)QMS_DATA.dpoints[i]);
            else
                printf ("    %s%g\n",
                        qmsStats[i].description, 
                        0.0);
        }
        else
        {
            printf ("    %s%d\n",  qmsStats[i].description,
                    (unsigned int)QMS_DATA.stats[i]);
        }
    }
    
    printf ("\n\n");
    
    qmsSessionPrintResults ();
    qmsSessionExit ();
    
    return (0);
}

